package org.hmn.jrsd;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.zip.CRC32;

import org.hmn.jrsd.rt.JrsdRtUtils;
import org.hmn.jrsd.rt.QColumn;

import antlr.RecognitionException;
import antlr.Token;

/*
 * 
 * Cree par Harry Karadimas a 30 juil. 2006
 */
/**
 * Represents the definition of a table, with descriptors for all associated
 * field, extra SQL, mappings, ...
 * Instanciated by the {@link JrsdParser}. 
 * @author karadimas
 */
public class TableDef
extends SqlElementDef
{
  String name;
  int version;
  List<ColumnDef> colDefs = new ArrayList<ColumnDef>();
  List<MappingDecl> mappingDecls = new ArrayList<MappingDecl>();
  String extraSql;
  HashMap<String, MappingDecl> mappingsByAlias =
    new HashMap<String, MappingDecl>();  
  JrsdSpec parent;  
  HashMap<String, ColumnDef> columnDefsByName = 
    new HashMap<String, ColumnDef>();
  ColumnDef pkDef;
  String postSql;
  Boolean quotingForced = null;
  
  public TableDef()
  {    
  }
    
  void addColumnDef(ColumnDef cd)
  {
    colDefs.add(cd);
    if (cd.isPrimaryKeyDecl()) {
      pkDef = cd;
      //update state of columndef
      for (Iterator<ColumnDef> iter = colDefs.iterator(); iter.hasNext();) {
        ColumnDef ccd = iter.next();
        ccd.setKey(isKey(ccd));
      }
    }
    else {
      columnDefsByName.put(cd.getName(), cd);
      cd.setKey(isKey(cd));
    }
  }
  
  public boolean isKey(ColumnDef cd)
  {
    if (pkDef == null) return false;
    return pkDef.containsKey(cd.getName());
    
  }
  void addMappingDecl(MappingDecl md)
  {
    md.setTableDef(this);
    mappingDecls.add(md);
    mappingsByAlias.put(md.getAlias(), md);
  }
  
  public List<ColumnDef> getColDefs()
  {
    return colDefs;
  }
  
  public void setColDefs(List<ColumnDef> colDefs)
  {
    this.colDefs = colDefs;
    if (colDefs != null) {
      for (Iterator<ColumnDef> iter = colDefs.iterator(); iter.hasNext();) {
        ColumnDef cd = iter.next();
        cd.setTableDef(this);
      }
    }
  }
  
  public String getName()
  {
    return name;
  }
  
  /**
   * Return name, quoted if necessary (q.i.n.).
   * @return
   */
  public String getQinName()
  {
    if (quotingForced != null) {
      if (quotingForced.booleanValue()) {
        //force usage of quotes
        return "\"" + name + "\"";
      }
      else {
        //forbid usage of quotes
        return name;
      }
    }
    return QColumn.quoteIfNecessary(name);
  }
  
  public void setName(String name)
  {
    this.name = name;
  }
  
  public void setName(Token t)
  {
    setName(t.getText());
  }
  
  public int getVersion()
  {
    return version;
  }
  
  public void setVersion(int version)
  {
    this.version = version;
  }
  
  public void setVersion(String v)
  {
    this.version = Integer.parseInt(v);
  }
  
  public void setVersion(Token t)
  {
    setVersion(t.getText());
  }
  
  /** return as array for stringtemplate */
  public ColumnDef[] getColumnDefinitions()
  {
    return (ColumnDef[]) colDefs.toArray(new ColumnDef[colDefs.size()]);    
  }
  
  public String getExtraSql()
  {
    return extraSql;
  }
  
  public void setExtraSql(String extraSql)
  {
    this.extraSql = extraSql;
  }
  
  public void setExtraSql(Token t)
  {
    setExtraSql(t == null ? null : t.getText());
  }
  
  public List<MappingDecl> getMappingDecls()
  {
    return mappingDecls;
  }
  
  public void setMappingDecls(List<MappingDecl> mappingDecls)
  {
    this.mappingDecls = mappingDecls;
  }
  
  public MappingDecl getMappingForAlias(String alias)
  {
    return (MappingDecl) mappingsByAlias.get(alias);  
  }
  
  public JrsdSpec getJrsdSpec()
  {
    return parent;
  }
  
  public void setJrsdSpec(JrsdSpec jrsdSpec)
  {
    this.parent = jrsdSpec;
  }
  
  public void addError(JrsdError e) { parent.addError(e); }
  
  /**
   * Calculate the checksum (aka digest) for this table. This number will 
   * uniquely identify a file and helps to detect unattended table changes.
   * @return
   */
  public long getChecksum()
  {
    CRC32 crc = new CRC32();
    if (literalSql == null) {
      JrsdRtUtils.update(crc, name);
      JrsdRtUtils.update(crc, String.valueOf(version));
    }
    else {
      JrsdRtUtils.update(crc, literalSql);
    }
    for (Iterator<ColumnDef> iter = colDefs.iterator(); iter.hasNext();) {
      ColumnDef cd = iter.next();    
      cd.update(crc);
    }
    JrsdRtUtils.update(crc, extraSql);
    return crc.getValue();
  }
  
  /**
   * @return The sql that is added at the end of the table's definition
   */
  public String getPostSql() {
    return postSql;
  }
  
  public void setPostSql(String postSql) {
    this.postSql = postSql;
  }
  
  public void setPostSql(Token t)
  {
    setPostSql(t == null ? null : t.getText());
  }

  /**
   * Add a directive. Supported directives: quote, unquote
   * @param flags
   * @throws RecognitionException
   */
  public void addDirectives(List<String> flags)
  throws RecognitionException
  {
    for (Iterator<String> iter = flags.iterator(); iter.hasNext();) {
      String directive = iter.next();
      if (directive.equals("quote")) {
        quotingForced = Boolean.TRUE;
      }
      else if (directive.equals("unquote")) {
        quotingForced = Boolean.FALSE;
      }
      else {
        throw new RecognitionException("Unknown table directive '"+directive+"'");
      }
    }//for    
  }

  public Boolean getQuotingForced() {
    return quotingForced;
  }

  public void setQuotingForced(Boolean quotingForced) {
    this.quotingForced = quotingForced;
  }  
  
}
