package org.hmn.jrsd.rt;

import java.io.Serializable;
import simpleorm.dataset.SFieldString;

/*
 * 
 * Cree par Harry Karadimas a 25 sept. 2006
 */

/**
 * A string that can not be longer than a maximum length, which constrasts
 * with java strings, that can grow as large as the memory allows. This
 * object carries its maximum length. There is no method to get the string 
 * value, toString() must be used for this. Note that the internal value can 
 * be null. Like "String", this object is immutable once it has been created.
 * To get a new FiniteString with the same maximum size as a given instance,
 * use {@link #makeNew(String)}. Ex :
 * <pre>
 *   FiniteString fs1 = new FiniteString(null, 23);
 *   FiniteString fs2 = fs1.makeNew("New value for fs2");
 * </pre> 
 * This object is not very convenient, (compared to {@link String}) but it
 * helps to transparently enforce correct maximum size and therefore avoids
 * database insert and update errors.
 * Note that here the string is silently truncated; this is not necessarily
 * what is desired (for example if the field is for the user's login, you should
 * report that the login is too long so the user knows). If you want exceptions
 * to be thrown, use regular Strings, and Simpleorm will detect too long strings
 * while validating the field's content.
 * Here is an example code to use it with String getter and setter, in
 * a Struts action object or any other bean object :
 * <pre>
 *   FiniteString myStringFs = new FiniteString(mySFieldString);
 *   ...
 *   public String getMyString() { return myStringFs.toString(); }
 *   public void setMyString(String val) { myStringFs = myStringFs.makeNew(val); }
 * </pre>
 * @author karadimas
 */
public class FiniteString
implements Serializable
{
  private static final long serialVersionUID = 1L;
  
  //immutable value, can be null
  private final String finalValue;
  private int maxLen; //maximum length allowed
  
  public static final String truncate(String val, int maxLen)
  {
    if (val == null) return null;
    if (val.length() <= maxLen) return val;
    else return val.substring(0, maxLen);    
  }
  
  /**
   * Initialize field with string value and maximum size.
   * @param val
   * @param maxLen
   */
  public FiniteString(String val, int maxLen) {
    this.maxLen = maxLen;
    finalValue = truncate(val, maxLen);    
  }

  /**
   * Initialize field with string value, take the maximum size from the
   * string field
   * @param val
   * @param fm
   */
  public FiniteString(String val, SFieldString fm) {
    this(val, fm.getMaxSize());
  }
  
  /**
   * Initialize field with the empty string (""), 
   * take the maximum size from string field
   * @param fm
   */
  public FiniteString(SFieldString fm) {
    this(null, fm.getMaxSize());
  }
  
  public String toString() { return finalValue; }
  
  public int getMaxLen() { return maxLen; }
  
  /**
   * Make a new FiniteString, that has the same max len as this one.
   * @param str
   * @return
   */
  public FiniteString makeNew(String str)
  {
    FiniteString fs = new FiniteString(str, maxLen);
    return fs;
  }
  
  /**
   * Test for equality. Only the strings are compared. FiniteString(s) can be
   * compared to another FiniteString; they are equal if their internal string
   * values are equal. A FiniteString can also be compared to a String. In this
   * case the String will be compared to the FiniteString's internal string.
   * The length of the string is not used for the comparison.
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    if (obj instanceof FiniteString) {
      FiniteString fs2 = (FiniteString) obj;
      if (finalValue == null) return fs2.finalValue == null;
      return finalValue.equals(fs2.finalValue);
    }
    if (obj instanceof String) {
      String str = (String) obj;
      return str.equals(finalValue);
    }
    return false;
  }
  
  @Override
  public int hashCode() {
    if (finalValue == null) return 0;
    return finalValue.hashCode();
  }

}
