package org.hmn.jrsd.rt;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/*
 * 
 * Cree par Harry Karadimas a 2 mai 08
 */
/**
 * Static utilities to call methods using reflection
 */
public class ReflectionHelpers
{

  public static void callVoid(String methodName, Object obj, Object arg)
  throws NoSuchMethodException, IllegalArgumentException, 
         IllegalAccessException, InvocationTargetException
  {
    Object[] args = { arg };
    callVoid(methodName, obj, args);
  }
  
  public static void callVoid(String methodName, Object obj, Object[] args)
  throws NoSuchMethodException, IllegalArgumentException, 
         IllegalAccessException, InvocationTargetException
  {
    Class<?>[] argClasses = new Class[args.length];
    for (int i = 0; i < argClasses.length; i++) {
      Object o = args[i];      
      argClasses[i] = (o == null) ? null : o.getClass();
    }
    Method m = obj.getClass().getMethod(methodName, argClasses);
    m.invoke(obj, args);
  }

  /**
   * Copy from a to b using reflection : for each "setXyz" method of B,
   * look for a "getXyz" method and call it. This allows copies between
   * objects that are mapped differently. The signature of getXyz and
   * setXyz must match exactly.
   * @param a Source object
   * @param b Destination object
   */
  public static void copyFromTo(Object a, Object b)
  {
    if (a == null) throw new NullPointerException("a");
    if (b == null) throw new NullPointerException("b");
    Class<?> ac = a.getClass();
    Class<?> bc = b.getClass();
    Method[] bms = bc.getMethods();
    for (int i = 0; i < bms.length; i++) {
      Method m = bms[i];
      String name = m.getName();
      //verify that the method starts with "set", and is followed by an
      //uppercase letter
      if (name.length() <= 3) continue;
      if (!name.startsWith("set")) continue;
      char c3 = name.charAt(3);
      if (c3 < 'A' | c3 > 'Z') continue;   
      //verify that setXyz only takes one parameter
      Class<?>[] paramTypes = m.getParameterTypes();
      if (paramTypes.length != 1) continue;
      String property = name.substring(3);
      Method getMeth = ReflectionHelpers.findMethod(ac, paramTypes[0], "get"+property);
      if (getMeth == null) continue;
      Object[] noArgs = {};
      try {
        Object val = getMeth.invoke(a, noArgs);
        Object[] setArgs = { val };
        m.invoke(b, setArgs);
      }
      catch (Exception e) {
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * Call a "get" method.
   * @param o
   * @param property The name of the property, case-sensitive (e.g. "foo" will
   *   call "getfoo", and "Foo" will call "getFoo")
   * @return
   */
  public static Object callGet(Object o, String property)
  {
    Class<?> oc = o.getClass();
    Object[] noArgs = {};
    try {
      Method getMeth = ReflectionHelpers.findFirstMethod(oc, "get"+property);
      if (getMeth == null) return null;
      Object val = getMeth.invoke(o, noArgs);
      return val;
    }
    catch (Exception e) {
      throw new RuntimeException(e);
    }
  }
  
  /** 
   * look for the method, but trap exception and return null if it doesn't exist 
   */
  public static Method findMethod(
      Class<?> objectClass, Class<?> returnType, String name)
  {
    Class<?>[] parameterTypes = {};
    try {
      //look for the method with the given type
      Method m = objectClass.getDeclaredMethod(name, parameterTypes);
      if (!returnType.equals(m.getReturnType())) return null;
      return m;
    }
    catch (NoSuchMethodException e) {
      return null;
    }    
  }

  public static Method findFirstMethod(Class<?> objCls, String methName)
  {
    Method[] ms = objCls.getMethods();
    for (int i = 0; i < ms.length; i++) {
      if (methName.equals(ms[i].getName())) {
        return ms[i];
      }
    }//for
    //not found, return null
    return null;
  }

  public static String capitalizeFirst(String str)
  {
    char[] ca = str.toCharArray();
    ca[0] = Character.toUpperCase(ca[0]);
    return new String(ca);
  }

  public static Method findSetMethod(String name, Class<?> objClass)
  {
    Class<?>[] args = { objClass };
    try {
      Method m = objClass.getMethod(name, args);
      return m;
    }
    catch (NoSuchMethodException e) {      
      return null;
    }
  }

  public static void setNullToAttribute(Object obj, String propName)
  {
    Class<?> objCls = obj.getClass();
    String mName = "set"+capitalizeFirst(propName);
    Method m = findFirstMethod(objCls, mName);
    if (m == null) {
      throw new NoSuchMethodError(mName);
    }
    Object[] args = { null };
    try {
      m.invoke(obj, args);
    }
    catch (Exception e) {
      throw new ValueTransferException(e);
    }
  }

  @SuppressWarnings("unchecked")
  public static void setValueToAttribute(Object obj, String propName, Object val)
  {
    Class<?> objCls = obj.getClass();
    String mName = "set"+capitalizeFirst(propName);
    if (TransportMappings.debugValueTransfer) {
      String valClass = "";
      if (val != null) valClass = ", class="+val.getClass();
      TransportMappings.lg.debug("Appel de setValueToAttribute pour "+mName+", val="+val+valClass);
    }
    //Class<?>[] pTypes = { val.getClass() };
    Object narrowedVal = val;
    try {
      Method m = findFirstMethod(objCls, mName);
      if (m == null) {
        throw new NoSuchMethodError("pas de methode "+mName+" dans "+objCls);
      }
      Class<?>[] paramTypes = m.getParameterTypes();
      if (paramTypes.length != 1) {
        throw new NoSuchMethodError("pas de methode "+mName+" compatible dans "+objCls);
      }
      Class<?> paramClass = paramTypes[0];
      if (val instanceof Number) {
        if (!(paramClass.isPrimitive() || Number.class.isAssignableFrom(paramClass))) {
          throw new NoSuchMethodError("pas de methode "+mName+" compatible avec" +
              " un argument de classe "+val.getClass()+" dans "+objCls);
        }
        //creer une instance compatible avec l'argument
        Class<? extends Number> paramClass2 = (Class<? extends Number>) paramClass;
        narrowedVal = ReflectionHelpers.narrowNumber((Number)val, paramClass2);
      }
      Object[] args = { narrowedVal };
      m.invoke(obj, args);
    }
    catch (Exception e) {
      TransportMappings.lg.error("(prop["+propName+"] val["+val+"])", e);
      throw new ValueTransferException(e);
    }
  }

  public static Number narrowNumber(Number nr, Class<? extends Number> numToClass)
  {
    if (numToClass.equals(Double.class) || numToClass.equals(double.class)) {
      return new Double(nr.doubleValue());
    }
    if (numToClass.equals(Float.class) || numToClass.equals(float.class)) {
      if (nr.doubleValue() > Float.MAX_VALUE || nr.doubleValue() < Float.MIN_VALUE) {
        throw new ValueTransferException("Can not store in a float : "+nr);
      }
      return new Float(nr.floatValue());
    } 
    if (numToClass.equals(Long.class) || numToClass.equals(long.class)) {
      return new Long(nr.longValue());
    } 
    if (numToClass.equals(Integer.class) || numToClass.equals(int.class)) {
      long lv = nr.longValue();
      if (Integer.MIN_VALUE <= lv && lv <= Integer.MAX_VALUE) {
        return new Integer(nr.intValue());
      }
      throw new ValueTransferException("Can not store in an int : "+nr);
    } 
    if (numToClass.equals(Short.class) || numToClass.equals(short.class)) {
      long lv = nr.longValue();
      if (Short.MIN_VALUE <= lv && lv <= Short.MAX_VALUE) {
        return new Short(nr.shortValue());
      }
      throw new ValueTransferException("Can not store in a short : "+nr);      
    } 
    if (numToClass.equals(Byte.class) || numToClass.equals(byte.class)) {
      long lv = nr.longValue();
      if (Byte.MIN_VALUE <= lv && lv <= Byte.MAX_VALUE) {
        return new Byte(nr.byteValue());
      }
      throw new ValueTransferException("Can not store in a byte : "+nr);      
    }
    throw new ValueTransferException("Unhandled target Number type : "+numToClass);        
  } 
  
}
