/*
 * Created on 23 mai 2005
 *
 */
package simpleorm.utils;

import java.util.HashMap;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import simpleorm.sessionjdbc.SSessionJdbc;

/**
 * A bridge from Simpleorm logging to log4j logging. This approach is different
 * than the approach proposed by simpleorm's white paper. All logging settings
 * are stored in an {@link Configuration} object, that can be cloned, and
 * set from properties. Each category of 
 * operations has its own logging level, that can be set independently of
 * others. The default levels are :
 * <ul>
 *   <li>connections : DEBUG
 *   <li>updates : DEBUG
 *   <li>queries : DEBUG
 *   <li>fields : DEBUG
 *   <li>debug : DEBUG
 *   <li>message : INFO
 *   <li>warn : WARN
 *   <li>error : ERROR
 *   <li>log : INFO
 * </ul>
 * You can retrieve an array with all the current levels, and use this array
 * to set the levels in another SLogLog4jBridge Object.
 * <br>
 * Similarly, you can use a properties file, to retrieve levels inside this
 * properties file.
 * <br>
 * You can also clone the 
 * @author KARADIMAS
 *
 */
public class SLogLog4jBridge 
extends SLog
implements Cloneable
{
  public static final String LN_ERROR = "error";
  public static final int LI_ERROR = 0;
  public static final String LN_CONNECTIONS = "connections";
  public static final int LI_CONNECTIONS = 1;
  public static final String LN_UPDATES = "updates";
  public static final int LI_UPDATES = 2;
  public static final String LN_QUERIES = "queries";
  public static final int LI_QUERIES = 3;
  public static final String LN_FIELDS = "fields";
  public static final int LI_FIELDS = 4;
  public static final String LN_DEBUG = "debug";
  public static final int LI_DEBUG = 5;
  public static final String LN_WARN = "warn";
  public static final int LI_WARN = 6;
  public static final String LN_LOG = "log";
  public static final int LI_LOG = 7;
  public static final String LN_MESSAGE = "message";
  public static final int LI_MESSAGE = 8;
 
  /**
   * Configuration value object holds the log levels for each simpleorm 
   * operation, and also
   * the logger to use. It's mainly just a value object, this is why the fields are
   * public. When cloning a Configuration, the logger is still the
   * same, and the logLevels array is cloned.
   * You instanciate a configuration and you set the current SLog's instance
   * field if it is a {@link SLogLog4jBridge} instance. To make this simpler,
   * the method {@link #applyTo(SSessionJdbc)} is provided.
   * @author 3073951
   *
   */
  public static class Configuration implements Cloneable {

    public Logger lg = Logger.getLogger(SLogLog4jBridge.class);
    public Level[] logLevels = {
        /* error */ Level.ERROR,  
        /* connections */ Level.DEBUG,
        /* updates */ Level.DEBUG,  
        /* queries */ Level.DEBUG,  
        /* fields */ Level.DEBUG,  
        /* debug */ Level.DEBUG,  
        /* warn */ Level.WARN,  
        /* log */ Level.INFO,
        /* message */ Level.INFO
    };
    
    @Override
    protected Object clone() {
      try {
        Configuration cln = (Configuration) super.clone();
        cln.logLevels = logLevels.clone();
        return cln;
      }
      catch (CloneNotSupportedException e) {
        throw new RuntimeException(e);
      }
    }
    
    /**
     * Apply this configuration to the given session. This is the most
     * used method.
     * @param s
     */
    public void applyTo(SSessionJdbc s) { applyTo(s.getLogger()); }
    
    public void applyTo(SLog s)
    {
      if (s instanceof SLogLog4jBridge) {
        SLogLog4jBridge s2 = (SLogLog4jBridge) s;
        s2.conf = this;
      }
    }
    
    public Level[] getLevels()
    {
      return logLevels;
    }
    
    /**
     * Use the given properties to set the levels, after having stripped the
     * given prefix.
     * Ex :
     * properties file :
     * <pre>
     * sloglog4.connections=OFF
     * sloglog4.queries=DEBUG
     * sloglog4.updates=INFO
     * </pre>
     * 
     * The call :
     * <pre>
     * setLevels(myProps, "sloglog4j.");
     * </pre>
     * 
     * @param props
     * @param prefix
     */
    public void setLevels(Properties props, String prefix)
    {
      for (Object key : props.keySet()) {
        String keyStr = String.valueOf(key);
        if (keyStr.startsWith(prefix)) {
          String propValue = props.getProperty(keyStr);
          String propName = keyStr.substring(prefix.length());
          setLevel(propName, propValue);
        }
      }// for
    }
    
    /**
     * Set levels using the level names that are in the map
     * <pre>
     * HashMap<String, String> lvls = new HashMap<String, String>();
     * lvls.put("debug", "INFO");
     * lvls.put("queries", "OFF");
     * setLevels(lvls);
     * </pre>
     * @param levelsByName
     */
    public void setLevels(HashMap<String, String> levelsByName)
    {
      Set<String> names = levelsByName.keySet();
      for (String name : names) {
        String levelStr = levelsByName.get(name);
        setLevel(name, levelStr);
      }
    }

    /**
     * Set the log levels from the given array
     * @param lvls
     */
    public void setLevels(Level[] lvls)
    {
      int ll = Math.min(lvls.length, logLevels.length);
      for (int i = 0; i < ll; i++) logLevels[i] = lvls[i];    
    }
    
    /**
     * Set an individual level, by name.
     * @param name
     * @param levelValue
     */
    public void setLevel(String nameAc, String levelValue)
    {
      String name = nameAc.toLowerCase();
      Integer ix = LEVELS_INDEXES_BY_NAME.get(name);
      if (ix >= 0) logLevels[ix] = Level.toLevel(levelValue);
    }
    
  }// Configuration
  
  public static final String[] LEVEL_NAMES =  {
    LN_ERROR, LN_CONNECTIONS, LN_UPDATES, LN_QUERIES, LN_FIELDS, LN_DEBUG,
    LN_WARN, LN_LOG, LN_MESSAGE    
  };
  
  private static final HashMap<String, Integer> LEVELS_INDEXES_BY_NAME = makeHmix();
  
  private static Configuration defaultConfiguration = new Configuration();
  
  private Configuration conf = defaultConfiguration;
  
  private static SLogLog4jBridge globalInstance = new SLogLog4jBridge();
  
  private static HashMap<String, Integer> makeHmix()
  {
    HashMap<String, Integer> hmix = new HashMap<String, Integer>();
    for (int i = 0; i < LEVEL_NAMES.length; i++) {
      hmix.put(LEVEL_NAMES[i], i);
    }
    return hmix;
  }
  
  @Override
  protected Object clone() {
    try {
      SLogLog4jBridge cloned = (SLogLog4jBridge) super.clone();
      return cloned;
    }
    catch (CloneNotSupportedException e) {
      throw new RuntimeException(e);
    }
  }
  
  public Logger getLogger() { return conf.lg; }

  public void setLogger(Logger logger) { conf.lg = logger; }
  
  /**
   * You normally don't need to use this constructor, is is used by newSLog()   
   */
  public SLogLog4jBridge() { }
  
  public SLogLog4jBridge(Logger lg) { 
    this.conf.lg = lg;
  }
  
  public SLogLog4jBridge(Configuration conf)
  {
    this.conf = conf;
  }
  
  /**
   * Call this method for system-wide SLog.log replacement. The logger that
   * is used is called "slog". Sets the globalInstance, which is also the
   * SLog.sessionlessLogger. If no logger named "slog" is found, the logger
   * associated with this class is used.
   * 
   */
  public static void useSLogLog4jBridge()
  {
    SLog.setSlogClass(SLogLog4jBridge.class);
    SLog.sessionlessLogger = newSLog();
    globalInstance = (SLogLog4jBridge)SLog.sessionlessLogger; 
    Logger slog = Logger.getLogger("slog");
    if (slog == null) slog = Logger.getLogger(SLogLog4jBridge.class);
    globalInstance.setLogger(slog);    
  }
  
  /**
   * @see simpleorm.core.SLog#error(java.lang.String)
   */
  @Override
  public void error(String msg) {
    conf.lg.log(conf.logLevels[LI_ERROR], "#### -ERROR " + sessionToString() + "_" + msg);
  }
    
  /**
   * @see simpleorm.core.SLog#warn(java.lang.String)
   */
  @Override
  public void warn(String msg) { 
    conf.lg.log(conf.logLevels[LI_WARN], "## -WARN " + sessionToString() + "_" + msg) ; 
  }
  
  // Open, begin, commit etc.
  @Override
  public void connections(String msg) {
    conf.lg.log(conf.logLevels[LI_CONNECTIONS], "  -" + msg);
  }

  @Override
  public void updates(String msg) {
    conf.lg.log(conf.logLevels[LI_UPDATES], "    -" + msg);
  }
  
  @Override
  public void queries(String msg) {
    conf.lg.log(conf.logLevels[LI_QUERIES], "      -" + msg);
  }
  
  @Override
  public void fields(String msg) {
    conf.lg.log(conf.logLevels[LI_FIELDS], msg);
  }
  
  /** 
   * always output
   * @see simpleorm.core.SLog#message(java.lang.String)
   */
  @Override
  public void message(String msg) { 
    conf.lg.log(conf.logLevels[LI_MESSAGE], " ---" + msg) ;
  }
  
  @Override
  public void debug(String msg) {
    conf.lg.log(conf.logLevels[LI_DEBUG], "          -" + "(" + msg + ")");
  }
  
  @Override
  protected void log(String msg) {
    conf.lg.log(conf.logLevels[LI_LOG], sessionToString() + msg);
  }
    
  public static SLogLog4jBridge getGlobalInstance() {
    return globalInstance;
  }

  public Level getConnectionsLevel() {
    return conf.logLevels[LI_CONNECTIONS];
  }

  public void setConnectionsLevel(Level connectionsLevel) {
    conf.logLevels[LI_CONNECTIONS] = connectionsLevel;
  }

  public Level getUpdatesLevel() {
    return conf.logLevels[LI_UPDATES];
  }

  public void setUpdatesLevel(Level updatesLevel) {
    conf.logLevels[LI_UPDATES] = updatesLevel;
  }

  public Level getQueriesLevel() {
    return conf.logLevels[LI_QUERIES];
  }

  public void setQueriesLevel(Level queriesLevel) {
    conf.logLevels[LI_QUERIES] = queriesLevel;
  }

  public Level getFieldsLevel() {
    return conf.logLevels[LI_FIELDS];
  }

  public void setFieldsLevel(Level fieldsLevel) {
    this.conf.logLevels[LI_FIELDS] = fieldsLevel;
  }

  public Level getDebugLevel() {
    return conf.logLevels[LI_DEBUG];
  }

  public void setDebugLevel(Level debugLevel) {
    conf.logLevels[LI_DEBUG] = debugLevel;
  }

  public Level getMessageLevel() {
    return conf.logLevels[LI_MESSAGE];
  }

  public void setMessageLevel(Level messageLevel) {
    conf.logLevels[LI_MESSAGE] = messageLevel;
  }

  public Level getWarnLevel() {
    return conf.logLevels[LI_WARN];
  }

  public void setWarnLevel(Level warnLevel) {
    conf.logLevels[LI_WARN] = warnLevel;
  }

  public Level getErrorLevel() {
    return conf.logLevels[LI_ERROR];
  }

  public void setErrorLevel(Level errorLevel) {
    conf.logLevels[LI_ERROR] = errorLevel;
  }

  public Level getLogLevel() {
    return conf.logLevels[LI_LOG];
  }

  public void setLogLevel(Level logLevel) {
    this.conf.logLevels[LI_LOG] = logLevel;
  }

  /**
   * Get the current configuration. There is no setter; you must use the
   * configuration object's "applyTo" method.
   * @return
   */
  public Configuration getConfiguration() { return conf; }
  
  public static Configuration getDefaultConfiguration() {
    return defaultConfiguration;
  }

  public static void setDefaultConfiguration(Configuration defaultConfiguration) {
    SLogLog4jBridge.defaultConfiguration = defaultConfiguration;
  }
  
}
